from typing import TYPE_CHECKING, Literal, Optional, Protocol, Type

from lonelypsp.stateful.messages.confirm_configure import B2S_ConfirmConfigure
from lonelypsp.stateless.make_strong_etag import StrongEtag


class IncomingAuthConfig(Protocol):
    async def setup_incoming_auth(self) -> None:
        """Prepares this authorization instance for use. If the incoming auth config
        is not re-entrant (i.e., it cannot be used by two clients simultaneously), it
        must detect this and error out.
        """

    async def teardown_incoming_auth(self) -> None:
        """Cleans up this authorization instance after use. This is called when a
        client is done using the auth config, and should release any resources it
        acquired during `setup_incoming_auth`.
        """

    async def is_receive_allowed(
        self,
        /,
        *,
        url: str,
        topic: bytes,
        message_sha512: bytes,
        now: float,
        authorization: Optional[str],
    ) -> Literal["ok", "unauthorized", "forbidden", "unavailable"]:
        """Determines if the given message can be received from the given topic. As
        we support very large messages, for authorization only the SHA-512 of
        the message should be used, which will be fully verified

        Args:
            url (str): the url the broadcaster used to reach us
            topic (bytes): the topic the message claims to be on
            message_sha512 (bytes): the sha512 of the message being received
            now (float): the current time in seconds since the epoch, as if from `time.time()`
            authorization (str, None): the authorization header they provided

        Returns:
            `ok`: if the message is allowed
            `unauthorized`: if the authorization header is required but not provided
            `forbidden`: if the authorization header is provided but invalid
            `unavailable`: if a service is required to check this isn't available.
              the message will be dropped.
        """

    async def is_missed_allowed(
        self,
        /,
        *,
        recovery: str,
        topic: bytes,
        now: float,
        authorization: Optional[str],
    ) -> Literal["ok", "unauthorized", "forbidden", "unavailable"]:
        """Determines if the indication that the subscriber may have missed a
        message from a broadcaster is allowed. This allows the subscriber to
        trigger a recovery mechanism to get back into a consistent state.

        Args:
            recovery (str): the url the missed message was sent to
            topic (bytes): the topic the message was on
            now (float): the current time in seconds since the epoch, as if from `time.time()`
            authorization (str, None): the authorization header they provided

        Returns:
            `ok`: if the message is allowed
            `unauthorized`: if the authorization header is required but not provided
            `forbidden`: if the authorization header is provided but invalid
            `unavailable`: if a service is required to check this isn't available
        """

    async def is_websocket_confirm_configure_allowed(
        self, /, *, message: B2S_ConfirmConfigure, now: float
    ) -> Literal["ok", "unauthorized", "forbidden", "unavailable"]:
        """Determines if the subscriber should continue with a websocket connection
        with a broadcaster who approved our configure message and responded back with
        the given confirm configure message.

        Args:
            message (B2S_ConfirmConfigure): the confirm configure message from the broadcaster
            now (float): the current time in seconds since the epoch, as if from `time.time()`

        Returns:
            `ok`: if the message is allowed
            `unauthorized`: if the authorization header is required but not provided
            `forbidden`: if the authorization header is provided but invalid
            `unavailable`: if a service is required to check this isn't available
        """


class OutgoingAuthConfig(Protocol):
    async def setup_outgoing_auth(self) -> None:
        """Prepares this authorization instance for use. If the outgoing auth config
        is not re-entrant (i.e., it cannot be used by two clients simultaneously), it
        must detect this and error out.
        """

    async def teardown_outgoing_auth(self) -> None:
        """Cleans up this authorization instance after use. This is called when a
        client is done using the auth config, and should release any resources it
        acquired during `setup_outgoing_auth`.
        """

    async def setup_subscribe_exact_authorization(
        self, /, *, url: str, recovery: Optional[str], exact: bytes, now: float
    ) -> Optional[str]:
        """Provides the authorization header that the subscriber should use to
        subscribe to a specific topic at the given url.

        Args:
            url (str): the url the subscriber is subscribing to
            recovery (str, None): the url that will receive MISSED messages for this
                subscription, if any
            exact (bytes): the exact topic they are subscribing to
            now (float): the current time in seconds since the epoch, as if from `time.time()`

        Returns:
            str, None: the authorization header to use, if any
        """

    async def setup_subscribe_glob_authorization(
        self, /, *, url: str, recovery: Optional[str], glob: str, now: float
    ) -> Optional[str]:
        """Provides the authoirzation header that the subscriber should use to subscribe
        to any topic that matches the given glob at the given url.

        Args:
            url (str): the url the subscriber is subscribing to
            recovery (str, None): the url that will receive MISSED messages for this
                subscription, if any
            glob (str): the glob pattern they are subscribing to
            now (float): the current time in seconds since the epoch, as if from `time.time()`

        Returns:
            str, None: the authorization header to use, if any
        """

    async def setup_notify_authorization(
        self, /, *, topic: bytes, message_sha512: bytes, now: float
    ) -> Optional[str]:
        """Provides the authorization header that the subscriber should use to
        ask a broadcaster to notify all subscribers to a topic about a message
        with the given hash. Only the hash of the message is used in authorization
        as the message itself may be very large; the hash will always be checked.

        Args:
            topic (bytes): the topic that the message is being sent to
            message_sha512 (bytes): the sha512 of the message being sent
            now (float): the current time in seconds since the epoch, as if from `time.time()`

        Returns:
            str, None: the authorization header to use, if any
        """

    async def setup_check_subscriptions_authorization(
        self, /, *, url: str, now: float
    ) -> Optional[str]:
        """Provides the authorization header that the subscriber should use to
        check the subscriptions that are currently active for the given url

        Args:
            url (str): the url the subscriber is checking
            now (float): the current time in seconds since the epoch, as if from `time.time()`

        Returns:
            str, None: the authorization header to use, if any
        """

    async def setup_set_subscriptions_authorization(
        self, /, *, url: str, strong_etag: StrongEtag, now: float
    ) -> Optional[str]:
        """Provides the authorization header that the subscriber should use to
        set the subscriptions that are currently active for the given url

        Unlike with the checking side which might compare the user being
        authenticated with vs the topics, there is generally no reason to need
        to view the specific globs/topics that are being subscribed to for
        generating the authorization token, as if they are not valid it will
        be caught by the broadcaster

        Args:
            url (str): the url the subscriber is setting
            strong_etag (StrongEtag): the strong etag of the subscriptions being set
            now (float): the current time in seconds since the epoch, as if from `time.time()`

        Returns:
            str, None: the authorization header to use, if any
        """

    async def setup_websocket_configure(
        self,
        /,
        *,
        subscriber_nonce: bytes,
        enable_zstd: bool,
        enable_training: bool,
        initial_dict: int,
    ) -> Optional[str]:
        """Provides the authorization header that the subscriber should use to
        configure the websocket connection with the broadcaster

        Args:
            subscriber_nonce (bytes): the 32 random bytes the subscriber is
                contributing toward the connection nonce
            enable_zstd (bool): whether to enable zstd compression
            enable_training (bool): whether to enable training mode
            initial_dict (int): the initial dictionary to use

        Returns:
            str, None: the authorization header to use, if any
        """


class AuthConfig(IncomingAuthConfig, OutgoingAuthConfig, Protocol): ...


class AuthConfigFromParts:
    """Convenience class to combine an incoming and outgoing auth config into an
    auth config
    """

    def __init__(self, incoming: IncomingAuthConfig, outgoing: OutgoingAuthConfig):
        self.incoming = incoming
        self.outgoing = outgoing

    async def setup_incoming_auth(self) -> None:
        await self.incoming.setup_incoming_auth()

    async def teardown_incoming_auth(self) -> None:
        await self.incoming.teardown_incoming_auth()

    async def setup_outgoing_auth(self) -> None:
        await self.outgoing.setup_outgoing_auth()

    async def teardown_outgoing_auth(self) -> None:
        await self.outgoing.teardown_outgoing_auth()

    async def is_receive_allowed(
        self,
        /,
        *,
        url: str,
        topic: bytes,
        message_sha512: bytes,
        now: float,
        authorization: Optional[str],
    ) -> Literal["ok", "unauthorized", "forbidden", "unavailable"]:
        return await self.incoming.is_receive_allowed(
            url=url,
            topic=topic,
            message_sha512=message_sha512,
            now=now,
            authorization=authorization,
        )

    async def is_missed_allowed(
        self,
        /,
        *,
        recovery: str,
        topic: bytes,
        now: float,
        authorization: Optional[str],
    ) -> Literal["ok", "unauthorized", "forbidden", "unavailable"]:
        return await self.incoming.is_missed_allowed(
            recovery=recovery,
            topic=topic,
            now=now,
            authorization=authorization,
        )

    async def is_websocket_confirm_configure_allowed(
        self, /, *, message: B2S_ConfirmConfigure, now: float
    ) -> Literal["ok", "unauthorized", "forbidden", "unavailable"]:
        return await self.incoming.is_websocket_confirm_configure_allowed(
            message=message,
            now=now,
        )

    async def setup_subscribe_exact_authorization(
        self, /, *, url: str, recovery: Optional[str], exact: bytes, now: float
    ) -> Optional[str]:
        return await self.outgoing.setup_subscribe_exact_authorization(
            url=url,
            recovery=recovery,
            exact=exact,
            now=now,
        )

    async def setup_subscribe_glob_authorization(
        self, /, *, url: str, recovery: Optional[str], glob: str, now: float
    ) -> Optional[str]:
        return await self.outgoing.setup_subscribe_glob_authorization(
            url=url,
            recovery=recovery,
            glob=glob,
            now=now,
        )

    async def setup_notify_authorization(
        self, /, *, topic: bytes, message_sha512: bytes, now: float
    ) -> Optional[str]:
        return await self.outgoing.setup_notify_authorization(
            topic=topic,
            message_sha512=message_sha512,
            now=now,
        )

    async def setup_check_subscriptions_authorization(
        self, /, *, url: str, now: float
    ) -> Optional[str]:
        return await self.outgoing.setup_check_subscriptions_authorization(
            url=url,
            now=now,
        )

    async def setup_set_subscriptions_authorization(
        self, /, *, url: str, strong_etag: StrongEtag, now: float
    ) -> Optional[str]:
        return await self.outgoing.setup_set_subscriptions_authorization(
            url=url,
            strong_etag=strong_etag,
            now=now,
        )

    async def setup_websocket_configure(
        self,
        /,
        *,
        subscriber_nonce: bytes,
        enable_zstd: bool,
        enable_training: bool,
        initial_dict: int,
    ) -> Optional[str]:
        return await self.outgoing.setup_websocket_configure(
            subscriber_nonce=subscriber_nonce,
            enable_zstd=enable_zstd,
            enable_training=enable_training,
            initial_dict=initial_dict,
        )


if TYPE_CHECKING:
    _: Type[AuthConfig] = AuthConfigFromParts



import asyncio

import uvicorn
from fastapi import APIRouter, FastAPI

from lonelypsc.config.http_config import (
    HttpPubSubBindManualConfig,
    HttpPubSubBindUvicornConfig,
)


class BindWithUvicornCallback:
    """Fulfills the HttpPubSubBindManualCallback using uvicorn as the runner"""

    def __init__(self, settings: HttpPubSubBindUvicornConfig):
        self.settings = settings

    async def __call__(self, router: APIRouter) -> None:
        app = FastAPI()
        app.include_router(router)
        app.router.redirect_slashes = False
        uv_config = uvicorn.Config(
            app,
            host=self.settings["host"],
            port=self.settings["port"],
            lifespan="off",
            # prevents spurious cancellation errors
            log_level="warning",
            # reduce default logging since this isn't the main deal for the process
        )
        uv_server = uvicorn.Server(uv_config)
        serve_task = asyncio.create_task(uv_server.serve())

        try:
            await asyncio.shield(serve_task)
        finally:
            uv_server.should_exit = True
            await serve_task


async def handle_bind_with_uvicorn(
    settings: HttpPubSubBindUvicornConfig,
) -> HttpPubSubBindManualConfig:
    """Converts the bind with uvicorn settings into the generic manual config"""
    return {
        "type": "manual",
        "callback": BindWithUvicornCallback(settings),
    }



import json
from typing import Literal, Optional, Tuple, cast

from lonelypsc.config.auth_config import IncomingAuthConfig, OutgoingAuthConfig
from lonelypsc.config.helpers.hmac_auth_config import (
    IncomingHmacAuth,
    IncomingHmacAuthSqliteDBConfig,
    OutgoingHmacAuth,
)
from lonelypsc.config.helpers.none_auth_config import (
    IncomingNoneAuth,
    OutgoingNoneAuth,
)
from lonelypsc.config.helpers.token_auth_config import (
    IncomingTokenAuth,
    OutgoingTokenAuth,
)


def get_auth_config_from_file(
    file_path: str,
) -> Tuple[IncomingAuthConfig, OutgoingAuthConfig]:
    """Reads the incoming/outgoing authorization specified in the file path,
    conventionally called `subscriber-secrets.json`, that was dumped
    from `httppubsubserver --setup`
    """
    with open(file_path, "r") as f:
        raw = json.load(f)

    if raw.get("version") != "1":
        raise ValueError(f"Unknown version {raw['version']}")

    incoming_type = cast(
        Literal["hmac", "token", "none"],
        "none" if "incoming" not in raw else raw["incoming"]["type"],
    )
    incoming_secret = cast(
        Optional[str], raw["incoming"]["secret"] if incoming_type != "none" else None
    )

    outgoing_type = cast(
        Literal["hmac", "token", "none"],
        "none" if "outgoing" not in raw else raw["outgoing"]["type"],
    )
    outgoing_secret = cast(
        Optional[str], raw["outgoing"]["secret"] if outgoing_type != "none" else None
    )

    incoming = cast(Optional[IncomingAuthConfig], None)
    outgoing = cast(Optional[OutgoingAuthConfig], None)

    if incoming_type == "none":
        incoming = IncomingNoneAuth()
    elif incoming_type == "token":
        assert incoming_secret is not None, "impossible"
        incoming = IncomingTokenAuth(incoming_secret)
    elif incoming_type == "hmac":
        assert incoming_secret is not None, "impossible"
        incoming = IncomingHmacAuth(
            incoming_secret, db_config=IncomingHmacAuthSqliteDBConfig(":memory:")
        )

    if outgoing_type == "none":
        outgoing = OutgoingNoneAuth()
    elif outgoing_type == "token":
        assert outgoing_secret is not None, "impossible"
        outgoing = OutgoingTokenAuth(outgoing_secret)
    elif outgoing_type == "hmac":
        assert outgoing_secret is not None, "impossible"
        outgoing = OutgoingHmacAuth(outgoing_secret)

    assert (
        incoming is not None
    ), f"unknown or unsupported incoming auth type {incoming_type}"
    assert (
        outgoing is not None
    ), f"unknown or unsupported outgoing auth type {outgoing_type}"

    return (incoming, outgoing)



import asyncio

from lonelypsc.ws.check_result import CheckResult
from lonelypsc.ws.state import SendingSimple, SendingState, StateOpen


def check_unsent_acks(state: StateOpen) -> CheckResult:
    """Checks if the left item from unsent acks can be moved to send_task;
    doing so if possible and nothing otherwise
    """
    if state.sending is not None:
        return CheckResult.CONTINUE

    if not state.unsent_acks:
        return CheckResult.CONTINUE

    state.sending = SendingSimple(
        type=SendingState.SIMPLE,
        task=asyncio.create_task(
            state.websocket.send_bytes(state.unsent_acks.popleft())
        ),
    )
    return CheckResult.RESTART



from typing import TYPE_CHECKING, Literal, Optional, Type

from lonelypsp.stateful.messages.confirm_configure import B2S_ConfirmConfigure
from lonelypsp.stateless.make_strong_etag import StrongEtag

from lonelypsc.config.auth_config import IncomingAuthConfig, OutgoingAuthConfig


class IncomingNoneAuth:
    """Implements the IncomingAuthConfig protocol with no-ops. Generally, use HMAC instead
    if you just want minimal setup, as it only requires syncing a single secret
    (hmac is still effective without https)
    """

    async def setup_incoming_auth(self) -> None: ...
    async def teardown_incoming_auth(self) -> None: ...

    async def is_receive_allowed(
        self,
        /,
        *,
        url: str,
        topic: bytes,
        message_sha512: bytes,
        now: float,
        authorization: Optional[str],
    ) -> Literal["ok", "unauthorized", "forbidden", "unavailable"]:
        return "ok"

    async def is_missed_allowed(
        self,
        /,
        *,
        recovery: str,
        topic: bytes,
        now: float,
        authorization: Optional[str],
    ) -> Literal["ok", "unauthorized", "forbidden", "unavailable"]:
        return "ok"

    async def is_websocket_confirm_configure_allowed(
        self, /, *, message: B2S_ConfirmConfigure, now: float
    ) -> Literal["ok", "unauthorized", "forbidden", "unavailable"]:
        return "ok"


class OutgoingNoneAuth:
    """Implements the OutgoingAuthConfig protocol with no-ops. Generally, use HMAC instead
    if you just want minimal setup, as it only requires syncing a single secret
    (hmac is still effective without https)
    """

    async def setup_outgoing_auth(self) -> None: ...
    async def teardown_outgoing_auth(self) -> None: ...

    async def setup_subscribe_exact_authorization(
        self, /, *, url: str, recovery: Optional[str], exact: bytes, now: float
    ) -> Optional[str]:
        return None

    async def setup_subscribe_glob_authorization(
        self, /, *, url: str, recovery: Optional[str], glob: str, now: float
    ) -> Optional[str]:
        return None

    async def setup_notify_authorization(
        self, /, *, topic: bytes, message_sha512: bytes, now: float
    ) -> Optional[str]:
        return None

    async def setup_check_subscriptions_authorization(
        self, /, *, url: str, now: float
    ) -> Optional[str]:
        return None

    async def setup_set_subscriptions_authorization(
        self, /, *, url: str, strong_etag: StrongEtag, now: float
    ) -> Optional[str]:
        return None

    async def setup_websocket_configure(
        self,
        /,
        *,
        subscriber_nonce: bytes,
        enable_zstd: bool,
        enable_training: bool,
        initial_dict: int,
    ) -> Optional[str]:
        return None


if TYPE_CHECKING:
    _: Type[IncomingAuthConfig] = IncomingNoneAuth
    __: Type[OutgoingAuthConfig] = OutgoingNoneAuth



import asyncio
import base64
import hmac
import math
import secrets
import sqlite3
import time
from typing import TYPE_CHECKING, Literal, Optional, Protocol, Tuple, Type, Union, cast

from lonelypsp.stateful.messages.confirm_configure import B2S_ConfirmConfigure
from lonelypsp.stateless.make_strong_etag import StrongEtag

from lonelypsc.config.auth_config import IncomingAuthConfig, OutgoingAuthConfig


class IncomingHmacAuthDBConfig(Protocol):
    async def setup_hmac_auth_db(self) -> None: ...
    async def teardown_hmac_auth_db(self) -> None: ...

    async def mark_token_used(
        self, /, *, token: bytes
    ) -> Literal["conflict", "ok"]: ...


class IncomingHmacAuthNoneDBConfig:
    """Does not store recent tokens and thus cannot check if they've been recently
    used. Technically, this is vulnerable to replay attacks, though the scope is
    rather limited.
    """

    async def setup_hmac_auth_db(self) -> None: ...
    async def teardown_hmac_auth_db(self) -> None: ...

    async def mark_token_used(self, /, *, token: bytes) -> Literal["conflict", "ok"]:
        return "ok"


if TYPE_CHECKING:
    _: Type[IncomingHmacAuthDBConfig] = IncomingHmacAuthNoneDBConfig


class IncomingHmacAuthSqliteDBConfig:
    """Stores recent tokens in a sqlite database, cleaning them in the background
    occassionally. On the subscriber side this is effective for preventing replay
    attacks assuming we only receive requests to our host as there is only one
    subscriber per host
    """

    def __init__(
        self,
        database: str,
        *,
        token_lifetime: int = 180,
        cleanup_batch_delay: float = 10.0,
    ) -> None:
        self.database = database
        """The database url. You can pass `:memory:` to create a SQLite database that
        exists only in memory, otherwise, this is typically the path to a sqlite file
        (usually has the `db` extension).
        """

        self.token_lifetime: int = token_lifetime
        """The minimum time in seconds before we forget about a token, must be at
        least as long as the token is accepted for this to be effective at preventing
        replay attacks.
        """

        self.cleanup_batch_delay: float = cleanup_batch_delay
        """The minimum time in seconds between cleaning up tokens that have expired"""

        self.conn: Optional[sqlite3.Connection] = None
        """The connection to the database"""

        self.cursor: Optional[sqlite3.Cursor] = None
        """The cursor that can be used iff you will be done using it before yielding
        to the event loop
        """

        self.background_task: Optional[asyncio.Task[None]] = None
        """The cleanup task that runs in the background"""

        self.cleanup_wakeup: asyncio.Event = asyncio.Event()
        """An event that is set whenever a token is created and will be waited on
        when there are no tokens in the store
        """

    async def setup_hmac_auth_db(self) -> None:
        assert self.background_task is None, "already entered, not re-entrant"
        conn = sqlite3.connect(self.database)
        try:
            cursor = conn.cursor()
            try:
                cursor.execute(
                    "CREATE TABLE IF NOT EXISTS httppubsub_hmac_auth_tokens (token BLOB PRIMARY KEY, expires_at INTEGER NOT NULL) WITHOUT ROWID"
                )
                cursor.execute(
                    "CREATE INDEX IF NOT EXISTS idx_httppubsub_hmac_auth_tokens_expires_at ON httppubsub_hmac_auth_tokens (expires_at)"
                )
                conn.commit()
                self.conn = conn
                self.cursor = cursor
                self.background_task = asyncio.create_task(self._cleanup_tokens())
            except BaseException:
                cursor.close()
                raise
        except BaseException:
            conn.close()
            raise

    async def teardown_hmac_auth_db(self) -> None:
        task = self.background_task
        conn = self.conn
        cursor = self.cursor
        self.background_task = None
        self.conn = None
        self.cursor = None
        try:
            if task is not None:
                task.cancel()
                await asyncio.wait([task])
        finally:
            try:
                if cursor is not None:
                    cursor.close()
            finally:
                if conn is not None:
                    conn.close()

    async def mark_token_used(self, /, *, token: bytes) -> Literal["conflict", "ok"]:
        assert self.conn is not None and self.cursor is not None, "not entered"
        self.cursor.execute(
            "SELECT 1 FROM httppubsub_hmac_auth_tokens WHERE token = ?",
            (token,),
        )
        result = self.cursor.fetchone() is not None
        if not result:
            self.cursor.execute(
                "INSERT INTO httppubsub_hmac_auth_tokens (token, expires_at) VALUES (?, ?)",
                (token, math.ceil(time.time() + self.token_lifetime)),
            )
        self.conn.commit()
        if not result:
            self.cleanup_wakeup.set()
        return "conflict" if result else "ok"

    async def _cleanup_tokens(self) -> None:
        assert self.conn is not None and self.cursor is not None, "not entered"
        conn = self.conn
        cursor = self.cursor

        while True:
            now = time.time()
            cursor.execute(
                "DELETE FROM httppubsub_hmac_auth_tokens WHERE expires_at < ?",
                (math.ceil(now),),
            )
            conn.commit()

            cursor.execute(
                "SELECT expires_at FROM httppubsub_hmac_auth_tokens ORDER BY expires_at ASC LIMIT 1"
            )
            next_expires_at = cast(Optional[int], cursor.fetchone())

            if next_expires_at is None:
                self.cleanup_wakeup.clear()
                await self.cleanup_wakeup.wait()
                await asyncio.sleep(self.token_lifetime + self.cleanup_batch_delay)
                continue

            await asyncio.sleep(max(next_expires_at - now, self.cleanup_batch_delay))


class IncomingHmacAuth:
    """Verifies that the authorization header is a recently generated, not
    recently used, HMAC token. These tokens can only be generated if the
    sender knows the shared secret.

    The authorization header is formatted as follows: `X-HMAC <timestamp>:<nonce>:<token>`,
    where timestamp is integer seconds from the epoch.

    This is a secure way to verify requests even if the underlying message
    and headers are not encrypted, as the shared secret cannot be discovered
    by an attacker who can only see the messages. However, it will not do
    anything to prevent the contents of the messages from being read

    Generally, this is an appropriate defense-in-depth measure to use in
    internal networks where you cannot setup TLS. It is also effective when TLS
    is available, though token authorization will require less CPU time for both
    broadcasters and subscribers and is secure if headers are encrypted.
    """

    def __init__(
        self,
        secret: str,
        *,
        token_lifetime: float = 120,
        db_config: IncomingHmacAuthDBConfig,
    ) -> None:
        self.secret = base64.urlsafe_b64decode(secret + "==")
        """The shared secret used to generate the HMAC tokens"""
        assert len(self.secret) == 64, "secret must be 64 bytes long"
        self.token_lifetime = token_lifetime
        """How long after a token is created that we still accept it"""
        self.db_config = db_config
        """The configuration for the database"""

    async def setup_incoming_auth(self) -> None:
        await self.db_config.setup_hmac_auth_db()

    async def teardown_incoming_auth(self) -> None:
        await self.db_config.teardown_hmac_auth_db()

    def _get_token(self, authorization: Optional[str], now: float) -> Union[
        Tuple[Literal["unauthorized", "forbidden"], None],
        Tuple[Literal["found"], Tuple[int, str, bytes]],
    ]:
        if authorization is None:
            return "unauthorized", None

        if not authorization.startswith("X-HMAC "):
            return "forbidden", None

        timestamp_nonce_and_token = authorization[len("X-HMAC ") :]
        sep_index = timestamp_nonce_and_token.find(":")
        if sep_index == -1:
            return "forbidden", None

        timestamp_str = timestamp_nonce_and_token[:sep_index]
        try:
            timestamp = int(timestamp_str)
        except ValueError:
            return "forbidden", None

        # clock drift means the time could be in the future
        if abs(now - timestamp) > self.token_lifetime:
            return "forbidden", None

        nonce_and_token = timestamp_nonce_and_token[sep_index + 1 :]
        sep_index = nonce_and_token.find(":")
        if sep_index == -1:
            return "forbidden", None

        nonce = nonce_and_token[:sep_index]

        hmac_token_str = nonce_and_token[sep_index + 1 :]
        try:
            hmac_token = base64.b64decode(hmac_token_str + "==")
        except ValueError:
            return "forbidden", None

        if len(hmac_token) != 64:
            return "forbidden", None

        return "found", (timestamp, nonce, hmac_token)

    async def _check_token(
        self, to_sign: bytes, hmac_token: bytes
    ) -> Literal["ok", "forbidden"]:
        expected_hmac = hmac.new(self.secret, to_sign, "sha512").digest()
        if not hmac.compare_digest(hmac_token, expected_hmac):
            return "forbidden"

        if await self.db_config.mark_token_used(token=hmac_token) == "conflict":
            return "forbidden"

        return "ok"

    async def is_receive_allowed(
        self,
        /,
        *,
        url: str,
        topic: bytes,
        message_sha512: bytes,
        now: float,
        authorization: Optional[str],
    ) -> Literal["ok", "unauthorized", "forbidden", "unavailable"]:
        assert len(message_sha512) == 64, "message_sha512 must be 64 bytes long"
        result = self._get_token(authorization, now)
        if result[0] != "found":
            return result[0]

        timestamp, nonce, hmac_token = result[1]
        encoded_url = url.encode("utf-8")
        encoded_timestamp = timestamp.to_bytes(8, "big")
        encoded_nonce = nonce.encode("utf-8")
        to_sign = b"".join(
            [
                encoded_timestamp,
                len(encoded_nonce).to_bytes(1, "big"),
                encoded_nonce,
                len(encoded_url).to_bytes(2, "big"),
                encoded_url,
                len(topic).to_bytes(2, "big"),
                topic,
                message_sha512,
            ]
        )
        return await self._check_token(to_sign, hmac_token)

    async def is_missed_allowed(
        self,
        /,
        *,
        recovery: str,
        topic: bytes,
        now: float,
        authorization: Optional[str],
    ) -> Literal["ok", "unauthorized", "forbidden", "unavailable"]:
        result = self._get_token(authorization, now)
        if result[0] != "found":
            return result[0]

        timestamp, nonce, hmac_token = result[1]
        encoded_recovery = recovery.encode("utf-8")
        encoded_timestamp = timestamp.to_bytes(8, "big")
        encoded_nonce = nonce.encode("utf-8")

        to_sign = b"".join(
            [
                encoded_timestamp,
                len(encoded_nonce).to_bytes(1, "big"),
                encoded_nonce,
                len(encoded_recovery).to_bytes(2, "big"),
                encoded_recovery,
                len(topic).to_bytes(2, "big"),
                topic,
            ]
        )
        return await self._check_token(to_sign, hmac_token)

    async def is_websocket_confirm_configure_allowed(
        self, /, *, message: B2S_ConfirmConfigure, now: float
    ) -> Literal["ok", "unauthorized", "forbidden", "unavailable"]:
        result = self._get_token(message.authorization, now)
        if result[0] != "found":
            return result[0]

        timestamp, nonce, hmac_token = result[1]
        encoded_timestamp = timestamp.to_bytes(8, "big")
        encoded_nonce = nonce.encode("utf-8")

        to_sign = b"".join(
            [
                encoded_timestamp,
                len(encoded_nonce).to_bytes(1, "big"),
                encoded_nonce,
                len(message.broadcaster_nonce).to_bytes(1, "big"),
                message.broadcaster_nonce,
            ]
        )
        return await self._check_token(to_sign, hmac_token)


class OutgoingHmacAuth:
    """Signs requests such that they can be verified by the subscriber but an
    eavesdropper cannot extract the secret.

    Specifically, the authorization header is formatted as follows:
    `X-HMAC <timestamp>:<nonce>:<token>`, where timestamp is integer seconds
    from the epoch, the nonce is to ensure uniqueness, and the token incorporates the
    relevant information and is signed with the shared secret.
    """

    def __init__(self, secret: str) -> None:
        self.secret = base64.urlsafe_b64decode(secret + "==")
        """The shared secret used to generate the HMAC tokens"""
        assert len(self.secret) == 64, "secret must be 64 bytes long"

    async def setup_outgoing_auth(self) -> None: ...
    async def teardown_outgoing_auth(self) -> None: ...

    def _make_nonce(self) -> str:
        return secrets.token_urlsafe(4)

    def _sign(self, to_sign: bytes, nonce: str, now: float) -> str:
        hmac_token = hmac.new(self.secret, to_sign, "sha512").digest()
        return (
            f"X-HMAC {int(now)}:{nonce}:{base64.b64encode(hmac_token).decode('ascii')}"
        )

    async def setup_subscribe_exact_authorization(
        self, /, *, url: str, recovery: Optional[str], exact: bytes, now: float
    ) -> Optional[str]:
        nonce = self._make_nonce()
        encoded_url = url.encode("utf-8")
        encoded_recovery = b"" if recovery is None else recovery.encode("utf-8")
        encoded_timestamp = int(now).to_bytes(8, "big")
        encoded_nonce = nonce.encode("utf-8")

        to_sign = b"".join(
            [
                encoded_timestamp,
                len(encoded_nonce).to_bytes(1, "big"),
                encoded_nonce,
                len(encoded_url).to_bytes(2, "big"),
                encoded_url,
                len(encoded_recovery).to_bytes(2, "big"),
                encoded_recovery,
                len(exact).to_bytes(2, "big"),
                exact,
            ]
        )
        return self._sign(to_sign, nonce, now)

    async def setup_subscribe_glob_authorization(
        self, /, *, url: str, recovery: Optional[str], glob: str, now: float
    ) -> Optional[str]:
        nonce = self._make_nonce()
        encoded_url = url.encode("utf-8")
        encoded_recovery = b"" if recovery is None else recovery.encode("utf-8")
        encoded_glob = glob.encode("utf-8")
        encoded_timestamp = int(now).to_bytes(8, "big")
        encoded_nonce = nonce.encode("utf-8")

        to_sign = b"".join(
            [
                encoded_timestamp,
                len(encoded_nonce).to_bytes(1, "big"),
                encoded_nonce,
                len(encoded_url).to_bytes(2, "big"),
                encoded_url,
                len(encoded_recovery).to_bytes(2, "big"),
                encoded_recovery,
                len(encoded_glob).to_bytes(2, "big"),
                encoded_glob,
            ]
        )
        return self._sign(to_sign, nonce, now)

    async def setup_notify_authorization(
        self, /, *, topic: bytes, message_sha512: bytes, now: float
    ) -> Optional[str]:
        assert len(message_sha512) == 64, "message_sha512 must be 64 bytes long"
        nonce = self._make_nonce()
        encoded_timestamp = int(now).to_bytes(8, "big")
        encoded_nonce = nonce.encode("utf-8")

        to_sign = b"".join(
            [
                encoded_timestamp,
                len(encoded_nonce).to_bytes(1, "big"),
                encoded_nonce,
                len(topic).to_bytes(2, "big"),
                topic,
                message_sha512,
            ]
        )
        return self._sign(to_sign, nonce, now)

    async def setup_check_subscriptions_authorization(
        self, /, *, url: str, now: float
    ) -> Optional[str]:
        nonce = self._make_nonce()
        encoded_url = url.encode("utf-8")
        encoded_timestamp = int(now).to_bytes(8, "big")
        encoded_nonce = nonce.encode("utf-8")

        to_sign = b"".join(
            [
                encoded_timestamp,
                len(encoded_nonce).to_bytes(1, "big"),
                encoded_nonce,
                len(encoded_url).to_bytes(2, "big"),
                encoded_url,
            ]
        )
        return self._sign(to_sign, nonce, now)

    async def setup_set_subscriptions_authorization(
        self, /, *, url: str, strong_etag: StrongEtag, now: float
    ) -> Optional[str]:
        nonce = self._make_nonce()
        encoded_url = url.encode("utf-8")
        encoded_etag = strong_etag.format.to_bytes(1, "big") + strong_etag.etag
        encoded_timestamp = int(now).to_bytes(8, "big")
        encoded_nonce = nonce.encode("utf-8")

        to_sign = b"".join(
            [
                encoded_timestamp,
                len(encoded_nonce).to_bytes(1, "big"),
                encoded_nonce,
                len(encoded_url).to_bytes(2, "big"),
                encoded_url,
                encoded_etag,
            ]
        )
        return self._sign(to_sign, nonce, now)

    async def setup_websocket_configure(
        self,
        /,
        *,
        subscriber_nonce: bytes,
        enable_zstd: bool,
        enable_training: bool,
        initial_dict: int,
    ) -> Optional[str]:
        nonce = self._make_nonce()
        encoded_nonce = nonce.encode("utf-8")
        encoded_timestamp = int(time.time()).to_bytes(8, "big")

        to_sign = b"".join(
            [
                encoded_timestamp,
                len(encoded_nonce).to_bytes(1, "big"),
                encoded_nonce,
                len(subscriber_nonce).to_bytes(1, "big"),
                subscriber_nonce,
                b"\1" if enable_zstd else b"\0",
                b"\1" if enable_training else b"\0",
                initial_dict.to_bytes(2, "big"),
            ]
        )
        return self._sign(to_sign, nonce, time.time())


if TYPE_CHECKING:
    __: Type[IncomingAuthConfig] = IncomingHmacAuth
    ___: Type[OutgoingAuthConfig] = OutgoingHmacAuth
